//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  ++39(49) 827-7216     FAX: ++39(49) 827-7102
//     EMail:  loreti@padova.infn.it
//     WWW:    http://wwwcdf.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------
//
//     See the header file "config.h" for more comments.
//
//     $Id: config.cxx,v 1.12 2003/07/04 09:53:50 loreti Exp $
//
// -------------------------------------------------------------

#include <algorithm>
#include <utility>
#include <cstdlib>
#include <cstring>
#include <climits>
#include "config.h"

using std::cerr;
using std::cout;
using std::ostream;
using std::strchr;
using std::string;
using std::vector;

// Local procedures

static char *getDelim(char ** pStart, char * sep);

// Methods for vEntry<T>

template <typename T> vEntry<T>::vEntry(
  const size_t       dim,
  const unsigned int flags,
  ...
) : _dim(dim), _flags(flags) {
  va_list vl;
  va_start(vl, flags);
  create(vl);
  va_end(vl);
}

template <typename T> vEntry<T>::vEntry(
  const size_t       dim,
  const unsigned int flags,
  va_list            vl
) : _dim(dim), _flags(flags) {
  create(vl);
}

template <typename T> void vEntry<T>::create(
  va_list vl
) {
  _targets = new vector<T *>;
  (* _targets).reserve( _dim );

  for (size_t i = 0;   i < _dim;   i++) {
    T * pT = va_arg(vl, T *);
    (* _targets).push_back(pT);
  }
}

template <typename T> void vEntry<T>::assign(
  const vector<T> * pValues
) {
  for (size_t i = 0;   i < _dim  &  i < pValues->size();   i++) {
    *((* _targets)[ i ]) = (* pValues)[ i ];
  }
}

template <typename T> void vEntry<T>::assign(
  const vector<T *> * pAddr
) {
  for (size_t i = 0;   i < _dim  &  i < pAddr->size();   i++) {
    *((* _targets)[ i ]) = *((* pAddr)[ i ]);
  }
}

template <typename T> void vEntry<T>::print(
  const string    name,
        ostream & os
) const {
  os << "Content of vEntry";
  if (!name.empty()) {
    os << " \"" << name << "\"";
  }
  os << " (" << _dim << " entr" << (_dim == 1 ? "y):\n" : "ies):\n");
  for (size_t i = 0;   i < _dim;   i++) {
    os << "  Element " << i << " has value \"" << *((* _targets)[ i ])
       << "\"\n";
  }
}

// Methods for cEntry

cEntry::cEntry(
  const cType    type,
  const size_t   dim,
  const unsigned flags,
  ...
) : _type(type) {
  va_list vl;
  va_start(vl, flags);
  create(dim, flags, vl);
  va_end(vl);
}

cEntry::cEntry(
  const cType    type,
  const size_t   dim,
  const unsigned flags,
  va_list        vl
) : _type(type) {
  create(dim, flags, vl);
}

cEntry::~cEntry()
{

// Converts "entry" from a (void *) to a (vEntry *), so that the
// destructor for * _entry is called

  switch (_type) {
    case C_LONG: {
      vEntry<long> * pE = static_cast<vEntry<long> *>( _entry );
      delete pE;
    } break;

    case C_ULONG: {
      vEntry<unsigned long> * pE =
        static_cast<vEntry<unsigned long> *>( _entry );
      delete pE;
    } break;

    case C_INT: {
      vEntry<int> * pE = static_cast<vEntry<int> *>( _entry );
      delete pE;
    } break;

    case C_UINT: {
      vEntry<unsigned int> * pE =
        static_cast<vEntry<unsigned int> *>( _entry );
      delete pE;
    } break;

    case C_DOUBLE: {
      vEntry<double> * pE = static_cast<vEntry<double> *>( _entry );
      delete pE;
    } break;

    case C_STRING: {
      vEntry<string> * pE = static_cast<vEntry<string> *>( _entry );
      delete pE;
    } break;
  }
}

void cEntry::create(
  const size_t   dim,
  const unsigned flags,
  va_list        vl
) {
  switch (_type) {
    case C_LONG: {
      vEntry<long> * pE = new vEntry<long>(dim, flags, vl);
      _entry = static_cast<void *>( pE );
    } break;

    case C_ULONG: {
      vEntry<unsigned long> * pE = new vEntry<unsigned long>(dim, flags, vl);
      _entry = static_cast<void *>( pE );
    } break;

    case C_INT: {
      vEntry<int> * pE = new vEntry<int>(dim, flags, vl);
      _entry = static_cast<void *>( pE );
    } break;

    case C_UINT: {
      vEntry<unsigned int> * pE = new vEntry<unsigned int>(dim, flags, vl);
      _entry = static_cast<void *>( pE );
    } break;

    case C_DOUBLE: {
      vEntry<double> * pE = new vEntry<double>(dim, flags, vl);
      _entry = static_cast<void *>( pE );
    } break;

    case C_STRING: {
      vEntry<string> * pE = new vEntry<string>(dim, flags, vl);
      _entry = static_cast<void *>( pE );
    }  break;
  }
}

void cEntry::print(
  const string    name,
        ostream & os
) const {
  if (!name.empty()) os << "\"" << name << "\" is a ";
  os << "cEntry of type <";
  switch (_type) {
    case C_LONG:
      os << "long>. ";
      {
        vEntry<long> * pE = static_cast<vEntry<long> *>( _entry );
        pE->print("");
      }
      break;

    case C_ULONG:
      os << "unsigned long>. ";
      {
        vEntry<unsigned long> * pE =
          static_cast<vEntry<unsigned long> *>( _entry );
        pE->print("");
      }
      break;

    case C_INT:
      os << "int>. ";
      {
        vEntry<int> * pE = static_cast<vEntry<int> *>( _entry );
        pE->print("");
      }
      break;

    case C_UINT:
      os << "unsigned int>. ";
      {
        vEntry<unsigned int> * pE =
          static_cast<vEntry<unsigned int> *>( _entry );
        pE->print("");
      }
      break;

    case C_DOUBLE:
      os << "double>. ";
      {
        vEntry<double> * pE = static_cast<vEntry<double> *>( _entry );
        pE->print("");
      }
      break;

    case C_STRING:
      os << "string>. ";
      {
        vEntry<string> * pE =
          static_cast<vEntry<string> *>( _entry );
        pE->print("");
      }
      break;
  }
}

cEntry::cRetv cEntry::assign(
  const vector<string *> * pStrings
) {
  cRetv retVal = C_OK;
  vector<string *>::size_type pSsize = pStrings->size();

  switch(_type) {

    case C_LONG: {
      vEntry<long> * pLong = static_cast<vEntry<long> *>( _entry );
      vector<long> * pValues = new vector<long>( pSsize );

      if (pLong->_dim != pSsize) {
        cerr << "cEntry::assign: warning: assigning " << pSsize
             << " values to  " << pLong->_dim << " variables\n";
        retVal = C_ERROR;
      }

      for (size_t i = 0;   i < pSsize;   i++) {
        char *pc;
        long value = strtol( ((* pStrings)[ i ])->c_str(), &pc, 0);

        if (!isspace(static_cast<unsigned char>(*pc))  &&  *pc != '\0') {
          cerr << "cEntry::assign: error converting string \""
               << *((* pStrings)[ i ]) << "\" to type 'long'\n";
          value = 0;
          retVal = C_ERROR;
        }
        (* pValues)[ i ] = value;
      }

      pLong->assign( pValues );
      delete pValues;
    } break;

    case C_ULONG: {
      vEntry<unsigned long> * pUlong =
        static_cast<vEntry<unsigned long> *>( _entry );
      vector<unsigned long> * pValues = new vector<unsigned long>( pSsize );

      if (pUlong->_dim != pSsize) {
        cerr << "cEntry::assign: warning: assigning " << pSsize
             << " values to  " << pUlong->_dim << " variables\n";
        retVal = C_ERROR;
      }

      for (size_t i = 0;   i < pSsize;   i++) {
        char *pc;
        unsigned long value = strtoul( ((* pStrings)[ i ])->c_str(), &pc, 0);
        if (!isspace(static_cast<unsigned char>(*pc))  &&  *pc != '\0') {
          cerr << "cEntry::assign: error converting string \""
               << *((* pStrings)[ i ]) << "\" to type 'unsigned long'\n";
          value = 0;
          retVal = C_ERROR;
        }
        (* pValues)[ i ] = value;
      }

      pUlong->assign( pValues );
      delete pValues;
    } break;

    case C_INT: {
      vEntry<int> * pInt = static_cast<vEntry<int> *>( _entry );
      vector<int> * pValues = new vector<int>( pSsize );

      if (pInt->_dim != pSsize) {
        cerr << "cEntry::assign: warning: assigning " << pSsize
             << " values to  " << pInt->_dim << " variables\n";
        retVal = C_ERROR;
      }

      for (size_t i = 0;   i < pSsize;   i++) {
        char *pc;
        long int value = strtol( ((* pStrings)[ i ])->c_str(), &pc, 0);

        if (!isspace(static_cast<unsigned char>(*pc))  &&  *pc != '\0') {
          cerr << "cEntry::assign: error converting string \""
               << *((* pStrings)[ i ]) << "\" to type 'int'\n";
          value = 0;
          retVal = C_ERROR;
        }
        if (value < INT_MIN   ||   value > INT_MAX) {
          cerr << "cEntry::assign: value \"" << *((* pStrings)[ i ])
               << "\" too large for an 'int'\n";
          value  = (value > 0 ? INT_MAX : INT_MIN);
          retVal = C_ERROR;
        }
        (* pValues)[ i ] = value;
      }

      pInt->assign( pValues );
      delete pValues;
    } break;

    case C_UINT: {
      vEntry<unsigned int> * pUint =
        static_cast<vEntry<unsigned int> *>( _entry );
      vector<unsigned int> * pValues = new vector<unsigned int>( pSsize );

      if (pUint->_dim != pSsize) {
        cerr << "cEntry::assign: warning: assigning " << pSsize
             << " values to  " << pUint->_dim << " variables\n";
        retVal = C_ERROR;
      }

      for (size_t i = 0;   i < pSsize;   i++) {
        char *pc;
        unsigned long value = strtoul( ((* pStrings)[ i ])->c_str(), &pc, 0);
        if (!isspace(static_cast<unsigned char>(*pc))  &&  *pc != '\0') {
          cerr << "cEntry::assign: error converting string \""
               << *((* pStrings)[ i ]) << "\" to type 'unsigned int'\n";
          value = 0;
          retVal = C_ERROR;
        }
        if (value > UINT_MAX) {
          cerr << "cEntry::assign: value \"" << *((* pStrings)[ i ])
               << "\" too large for an 'unsigned int'\n";
          value  = UINT_MAX;
          retVal = C_ERROR;
        }
        (* pValues)[ i ] = value;
      }

      pUint->assign( pValues );
      delete pValues;
    } break;

    case C_DOUBLE: {
      vEntry<double> *pDouble = static_cast<vEntry<double> *>(_entry);
      vector<double> * pValues = new vector<double>( pSsize );

      if (pDouble->_dim != pSsize) {
        cerr << "cEntry::assign: warning: assigning " << pSsize
             << " values to  " << pDouble->_dim << " variables\n";
        retVal = C_ERROR;
      }

      for (size_t i = 0;   i < pSsize;   i++) {
        char *pc;
        double value = strtod( ((* pStrings)[ i ])->c_str(), &pc);
        if (!isspace(static_cast<unsigned char>(*pc))  &&  *pc != '\0') {
          cerr << "cEntry::assign: error converting string \""
               << *((* pStrings)[ i ]) << "\" to type 'double'\n";
          value = 0.0;
          retVal = C_ERROR;
        }
        (* pValues)[ i ] = value;
      }

      pDouble->assign( pValues );
      delete pValues;
    } break;

    case C_STRING: {
      vEntry<string> *pS = static_cast<vEntry<string> *>(_entry);

      if (pS->_dim != pSsize) {
        cerr << "cEntry::assign: warning: assigning " << pSsize
             << " values to  " << pS->_dim << " variables\n";
        retVal = C_ERROR;
      }

      pS->assign( pStrings );
    } break;
  }

  return retVal;
}

// Methods for config

const size_t config::lineLength = 128;

void config::addConfig(
  const string        keyword,
  const cEntry::cType type,
  const size_t        dim,
  const unsigned      flags,
  ...
) {
  va_list vl;
  va_start(vl, flags);
  cEntry * pCE = new cEntry(type, dim, flags, vl);
  va_end(vl);

  std::pair<configDict::iterator, bool> it;

  it = _dBase->insert( std::make_pair(keyword, pCE) );
  if (!it.second) {
    cerr << "config::addConfig: keyword \"" << keyword
         << "\" is already known; ignored.\n";
  }
}

config::~config()
{
  for_each(_dBase->begin(), _dBase->end(), delPtrConfig());
  delete _dBase;
}

void config::doConfig(
  const string           * pKeyword,
  const vector<string *> * pStrings,
  const string             fileName,
  const size_t             lineNr
) {
  bool err = false;
  configDict::iterator it;

  it = _dBase->find(*pKeyword);
  if (it == _dBase->end()) {
    cerr << "config::doConfig: keyword \"" << *pKeyword
         << "\" is unknown\n";
    err = true;
  } else {
    cEntry * pCE = it->second;
    if (pCE->assign( pStrings ) == cEntry::C_ERROR) err = true;
  }

  if (err) {
    cerr << "\t(while reading file \"" << fileName
         << "\", line " << lineNr << ")\n";
  }
}

void config::print(
  const string    name,
        ostream & os
) const {
  os << "* Content of config instantiation \"" << name << "\":\n";

  for_each(_dBase->begin(), _dBase->end(), printConf(os));
}

void config::readConfig(
  std::ifstream * inFile,
  const string    name
) {
  char buffer[lineLength];
  size_t lineNr = 0;

  vector<string *> * pStrings = new vector<string *>;

  while (inFile->getline(buffer, lineLength)) {
    ++lineNr;

// For every line, strips the comments inserting a '\0' over the first
// unquoted ';' found.

    char *pc, *pName, *pValue;

    for (pc = buffer,  pName = buffer + 1;   *pc;   ) {
      if (*pc == '\\'   &&   *pName == ';') {
        std::strcpy(pc, pName);
      } else if (*pc == ';') {
        *pc = '\0';
        break;
      }
      pc = pName++;
    }

// Skips the leading whitespace; "pName" points to the first non blank
// character (or to the trailing '\0': an empty line).  The first non
// blank character cannot be a '='.

    for (pName = buffer;
         *pName && isspace(static_cast<unsigned char>(*pName));
         pName++) { }

    if (*pName == '\0') {
      continue;
    } else if (*pName == '=') {
      cerr
        << "config::readConfig: variable name missing in an assignment\n"
        << "\t(while reading file \"" << name
        << "\", line " << lineNr << ")\n";
      continue;
    }

// Looks for the assignment operator '='; delimits the symbolic name
// field.

    if ((pValue = strchr(pName, '=')) == 0) {
      for (pc = pName;
           *pc && !isspace(static_cast<unsigned char>(*pc));
           pc++) { }
      *pc = '\0';
      cerr
        << "config::readConfig: no assignment to keyword \"" << pName << "\"\n"
        << "\t(while reading file \"" << name
        << "\", line " << lineNr << ")\n";
      continue;
    }

    *pValue++ = '\0';
    for (pc = pName;
         *pc && !isspace(static_cast<unsigned char>(*pc));
         pc++) { }
    *pc = '\0';

// Skips the leading whitespace after the '=', to find the beginning
// of the "value" field

    while (*pValue && isspace(static_cast<unsigned char>(*pValue))) ++pValue;
    if (*pValue == 0) {
      cerr
        << "config::readConfig: no value for variable \"" << pName << "\"\n"
        << "\t(while reading file \"" << name
        << "\", line " << lineNr << ")\n";
      continue;
    }

// Parses the "value" field in substrings according to the ':'
// separators

    string keyword( pName );

    while ((pc = getDelim(&pValue, ":")) != 0) {
      string * pField = new string( pc );
      pStrings->push_back( pField );
    }

// Calls doConfig, then cleans up and goes to the next line

    doConfig(&keyword, pStrings, name, lineNr);

    for_each(pStrings->begin(), pStrings->end(), delSP());
    pStrings->clear();
  }

  if (!inFile->eof()) {
    cerr << "config::readConfig: file \"" << name
         << "\", read error after line " << lineNr << std::endl;
  }
  delete pStrings;
}

// Local procedures

static char *getDelim(
  char **pStart,
  char  *sep
) {
  register char *retVal;
  register char *pc;
  register char *fb = 0;

  if (!pStart  ||  !(retVal = *pStart)  ||  !sep) return 0;

  while (*retVal  &&  isspace(static_cast<unsigned char>(*retVal))) retVal++;

  if (!*(pc = retVal)) return 0;

  while (*pc) {
    if (isspace(static_cast<unsigned char>(*pc))) {
      if (!fb) fb = pc;
    } else {
      if (strchr(sep, *pc)) {
        *pc++ = '\0';
        break;
      }
      fb = 0;
    }
    pc++;
  }

  if (fb) *fb = '\0';

  *pStart = pc;

  return retVal;
}
