//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  ++39(49) 827-7216     FAX: ++39(49) 827-7102
//     EMail:  loreti@padova.infn.it
//     WWW:    http://wwwcdf.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------
//
//     Header class for config.cxx
//
//     $Id: config.h,v 1.8 2001/08/16 18:14:22 loreti Exp $
//
// -------------------------------------------------------------

#ifndef CONFIG_H_
#define CONFIG_H_

#include <fstream>
#include <functional>
#include <iostream>
#include <stdexcept>
#include <string>
#include <vector>
#include <map>
#include <cstdarg>

// A template class to hold pointers to a fixed number of variables.
//    _dim:     number of variables;
//    _targets: a vector of pointers.
//
// Constructors are provided to create such a class from a list of
// arguments, or from a generic pointer (a "va_list") to a list of
// arguments.  Two private methods, "create" and "assign" are also
// supplied: the first is internally used by the two overloaded
// constructors, and the second is used by the "assign" method of the
// friend class "cEntry" to store new values into the pointed
// variables (will be called with a vector of _values_ for numerical
// variables, with a vector of _addresses_ for string variables).

template <typename T> class vEntry {
  friend class cEntry;

private:
  size_t             _dim;
  unsigned int       _flags;
  std::vector<T *> * _targets;

  void create(va_list vl);
  void assign(const std::vector<T>   * pValues);
  void assign(const std::vector<T *> * pAddr);

  // Prevent use of the copy constructor and of the assignment operator

  const vEntry & operator = (const vEntry & rhs);
        vEntry              (const vEntry & rhs);

public:
  vEntry(const size_t dim, const unsigned flags, ...);
  vEntry(const size_t dim, const unsigned flags, va_list vl);
  ~vEntry() { delete _targets; }
  void print(const std::string name, std::ostream & os = std::cout) const;
};

// A class to store a pointer to a generic "vEntry", with the type of
// the pointed variables.
//
// Constructors are provided to create such a class from a list of
// arguments, or from a generic pointer (a "va_list") to a list of
// arguments.  Two private methods, "create" and "assign" are also
// supplied: the first is internally used by the two overloaded
// constructors, and the second is used by the "doConfig" method of
// the friend class "config" to store new values into the variables
// pointed to by the contained vEntry.
//
// The public enumeration "cEntry::cType" supplies a list of the
// allowed types; cEntry::cRetv the possible return values from
// cEntry::assign() .

class cEntry {
  friend class config;

public:
  enum cType { C_LONG, C_ULONG, C_INT, C_UINT, C_DOUBLE, C_STRING };
  enum cRetv { C_ERROR = 1, C_OK };

private:
  cType   _type;
  void  * _entry;

  void create(const size_t dim, const unsigned flags, va_list vl);
  cRetv assign(const std::vector<std::string *> * pStrings);

  // Prevent use of the copy constructor and of the assignment operator

  const cEntry & operator = (const cEntry & rhs);
        cEntry              (const cEntry & rhs);

public:
  cEntry(const cType type, const size_t dim, const unsigned flags, ...);
  cEntry(const cType type, const size_t dim, const unsigned flags, va_list vl);
  ~cEntry();
  void print(const std::string name, std::ostream & os = std::cout) const;
};

// A class to store correspondences between strings (keywords) and
// vectors of pointers to generic variables ("cEntry" instances).
//
// The private method "doConfig" is used by readConfig to store values
// into the pointed variables.  The two public methods "addConfig" and
// "readConfig" are used: the first to add a new keyword to the
// database; and the second to read a file looking for assignments.
//
// The public constant "config::lineLength" fixes the maximum length
// of an input line.

typedef std::map<std::string, cEntry *> configDict;

class config {
private:
  configDict * _dBase;

  void doConfig(const std::string * pKeyword,
                const std::vector<std::string *> * pStrings,
                const std::string fileName,
                const size_t lineNr);

  // Prevent use of the copy constructor and of the assignment operator

  const config & operator = (const config & rhs);
        config              (const config & rhs);

public:
  config() { _dBase = new configDict; }
  ~config();

  static const size_t lineLength;

  void addConfig(const std::string keyword, const cEntry::cType type,
                 const size_t dim, const unsigned flags, ...);
  void readConfig(std::ifstream * inFile, const std::string name);
  void print(const std::string name, std::ostream & os = std::cout) const;
};

// A function object used to delete pointers to cEntry, in the class
// config dtor

typedef std::pair<const std::string, cEntry *> configEntry;

struct delPtrConfig : public std::unary_function<configEntry &, void> {
  void operator () (configEntry & ce) {
    delete ce.second;
  }
};

// A function object used from config::print() to output its content

class printConf : public std::unary_function<configEntry &, void> {
  std::ostream & os;

public:
  printConf(std::ostream & o) : os(o) {}
  void operator () (configEntry & ce) {
    os << "Keyword: \"" << ce.first << "\";\n";
    ce.second->print("");
  }
};

// A function object used by config::readConfig for internal cleanup

struct delSP : public std::unary_function<std::string *, void> {
  void operator () (std::string * pS) { delete pS; }
};

#endif // CONFIG_H_
