//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// getDelim is a procedure that parses a string in tokens according
// with embedded separator characters defined by the user.  A simple
// test program is enclosed; see below for the syntax of getDelim and
// for the details of the implemented algorithm.
//
// $Id: getDelim.cxx,v 1.5 2005/01/21 08:29:43 loreti Exp $
//
// -------------------------------------------------------------------

#include "getDelim.hh"

using std::auto_ptr;
using std::string;
using std::vector;

#ifdef TEST
#include <iostream>

using std::cin;
using std::cout;
using std::getline;

int main() {                    // Quick and dirty.  Sorry...
  for (;;) {
    cout << "String (\"\\\" to exit) ? ";
    string s;
    getline(cin, s);
    if (s == "\\") return 0;

    cout << "Delimiters ? ";
    string d;
    getline(cin, d);
    if (d.find_first_of(" \t\f\n\r\v") != string::npos) {
      cout << "You cannot embed whitespace in the delimiters!\n";
      continue;
    }

    auto_ptr< vector<string> > pvs = getDelim(s, d);

    cout << "String: \"" << s << "\"; "
         << "delimiters: \"" << d << "\"; "
         << pvs->size() << " tokens parsed.\n";

    if (!pvs->empty()) {
      vector<string>::size_type      n;
      vector<string>::const_iterator it;
      vector<string>::const_iterator itEnd = pvs->end();

      for (n = 1,  it = pvs->begin();   it != itEnd;   ++n,  ++it) {
        cout << n << ":\t\"" << *it << "\"\n";
      }
    }
  }
}
#endif // TEST

// -------------------------------------------------------------------
// A call to getDelim() returns an auto_ptr to a vector of strings,
// containing all the "tokens" extracted from another given string
// according to a set of "separator" characters.  A "token" is defined
// in the following way: the original string is first divided in
// "fields" limited by the start of the string, the end of the string,
// and every character chosen as "separator"; the sequence from the
// first to the last non-whitespace character in every "field" is a
// "token".
//
// "Null tokens", i.e. nothing and/or whitespace only between two
// separators (or between the beginning of the string and a separator,
// or between a separator and the end of the string), are allowed; in
// that case the returned vector contains an empty string.
//
// getDelim() is implemented as a finite state machine; the SM driver
// table is the following:
//
//                +----------------+----------------+----------------+
//                |  end of string |    delimiter   | non-whitespace |
// ---------------+----------------+----------------+----------------+
// state 0        |      exit      |    push null   | set start, end |
// (initial)      |                | token; state=1 |    state = 2   |
// ---------------+----------------+----------------+----------------+
// state 1        |    push null   |    push null   | set start, end |
// (out of token) |   token; exit  | token; state=1 |    state = 2   |
// ---------------+----------------+----------------+----------------+
// state 2        |   push token;  |   push token;  |   set end of   |
// (inside token) |     exit       |    state = 1   | token; state=2 |
// ---------------+----------------+----------------+----------------+
//
// However what is saved is not start and end of the token, but start
// and _length_; and there is not an explicit 'state' variable.  state
// is 0 when the token length is 0 and no tokens have been pushed;
// state is 1 when the length is 0 and some tokens have been pushed;
// and state is 2 when the token length is positive.
// -------------------------------------------------------------------

auto_ptr< vector<string> > getDelim(
  const string & in,            // String to be parsed
  const string & sep            // Separator characters
) {
  auto_ptr< vector<string> > ret( new vector<string> );
  string::size_type i           = 0;
  string::size_type tokenStart  = 0;
  string::size_type tokenLength = 0;

  while (i < in.size()) {
    register char c( in[i++] );

    if (sep.find(c) != string::npos) {
      ret->push_back( in.substr(tokenStart, tokenLength) );
      tokenStart = tokenLength = 0;
    } else if (!isspace(static_cast<unsigned char>(c))) {
      if (tokenLength == 0) tokenStart = i - 1;
      tokenLength = i - tokenStart;
    }
  }

  if (tokenLength > 0   ||   !ret->empty()) {
    ret->push_back( in.substr(tokenStart, tokenLength) );
  }

  return ret;
}
