//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
//
// getToken is a procedure that parses a string in tokens according
// with the embedded whitespace and with separator characters defined
// by the user.  A simple test program is enclosed; see below for the
// syntax of getToken and for the details of the implemented
// algorithm.
//
// $Id: getToken.cxx,v 1.5 2005/01/21 08:29:43 loreti Exp $
//
// -------------------------------------------------------------------

#include "getToken.hh"

using std::auto_ptr;
using std::string;
using std::vector;

auto_ptr< vector<string> > getToken(const string &, const string &);

#ifdef TEST
#include <iostream>

using std::cin;
using std::cout;
using std::getline;

int main() {                    // Quick and dirty.  Sorry...
  for (;;) {
    cout << "String (\"\\\" to exit) ? ";
    string s;
    getline(cin, s);
    if (s == "\\") return 0;

    cout << "Delimiters ? ";
    string d;
    getline(cin, d);
    if (d.find_first_of(" \t\f\n\r\v") != string::npos) {
      cout << "You cannot embed whitespace in the delimiters!\n";
      continue;
    }

    auto_ptr< vector<string> > pvs = getToken(s, d);

    cout << "String: \"" << s << "\"; "
         << "delimiters: \"" << d << "\"; "
         << pvs->size() << " tokens parsed.\n";

    if (!pvs->empty()) {
      vector<string>::size_type      n;
      vector<string>::const_iterator it;
      vector<string>::const_iterator itEnd = pvs->end();

      for (n = 1,  it = pvs->begin();   it != itEnd;   ++n,  ++it) {
        cout << n << ":\t\"" << *it << "\"\n";
      }
    }
  }
}
#endif // TEST

// -------------------------------------------------------------------
// A call to getToken() returns an auto_ptr to a vector of strings,
// containing all the "tokens" extracted from another given string
// according to a set (that may be empty) of "separator" characters
// supplied by the user and to whitespace.  A "token" is defined in
// the following way: the original string is first divided in "fields"
// limited by the start of the string, the end of the string, the
// embedded whitespace, and every character chosen as "separator"; the
// sequence of non-whitespace character in every "field" is a "token".
//
// "Null tokens", i.e. nothing and/or whitespace only between two
// separators (or between the beginning of the string and a separator,
// or between a separator and the end of the string), are allowed; in
// that case the returned vector contains an empty string.
//
// getToken() is implemented as a finite state machine; the SM driver
// table is the following:
//
//                +-------------+---------------+------------+---------------+
//                |  end string |   delimiter   | whitespace |     other     |
// ---------------+-------------+---------------+------------+---------------+
// state 0        |    exit     | push null to- |  continue  | set start,end |
// (initial)      |             | ken; state=1  |            |   state = 2   |
// ---------------+-------------+---------------+------------+---------------+
// state 1        |  push null  | push null to- |  continue  | set start,end |
// (out of token) | token; exit | ken; state=1  |            |   state = 2   |
// ---------------+-------------+---------------+------------+---------------+
// state 2        | push token; |  push token;  | push token |    set end    |
// (inside token) |    exit     |   state = 1   | state = 3  |   state = 2   |
// ---------------+-------------+---------------+------------+---------------+
// state 3 (after |    exit     |   state = 1   |  continue  | set start,end |
// end of token)  |             |               |            |   state = 2   |
// ---------------+-------------+---------------+------------+---------------+
//
// However what is saved is not start and end of the token, but start
// and _length_ of the token.
// -------------------------------------------------------------------

auto_ptr< vector<string> > getToken(
  const string & in,            // String to be parsed
  const string & sep            // Separator characters
) {
  auto_ptr< vector<string> > ret( new vector<string> );
  int state                     = 0;
  string::size_type i           = 0;
  string::size_type tokenStart  = 0;
  string::size_type tokenLength = 0;

  while (i < in.size()) {
    register char c( in[i++] );

    if (!sep.empty()   &&   sep.find(c) != string::npos) {
      if (state != 3) {
        ret->push_back( in.substr(tokenStart, tokenLength) );
        tokenStart = tokenLength = 0;
      }
      state = 1;
    } else if (isspace(static_cast<unsigned char>(c))) {
      if (state == 2) {
        ret->push_back( in.substr(tokenStart, tokenLength) );
        tokenStart = tokenLength = 0;
        state      = 3;
      }
    } else {
      if (state != 2) {
        tokenStart = i - 1;
        state      = 2;
      }
      tokenLength = i - tokenStart;
    }
  }

  if (state == 1   ||   state == 2) {
    ret->push_back( in.substr(tokenStart, tokenLength) );
  }

  return ret;
}
