/*------------------------------------------------------*
 | Author: Maurizio Loreti, aka MLO or (HAM) I3NOO      |
 | Work:   University of Padova - Department of Physics |
 |         Via F. Marzolo, 8 - 35131 PADOVA - Italy     |
 | Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102 |
 | EMail:  loreti@padova.infn.it                        |
 | WWW:    http://wwwcdf.pd.infn.it/~loreti/mlo.html    |
 *------------------------------------------------------*

 A call to getToken() parses a C string in a collection of "tokens".
 A "token" is defined in the following way: the original string is
 first divided in "fields" limited by the start of the string, the end
 of the string, the contained whitespace and every character chosen as
 "separator"; the sequence of non-whitespace character in every
 "field" is a "token".

 The integer value returned from getToken() is the number of parsed
 tokens, or a negative value in case of error (GETTOKEN_WRONG = -1 for
 wrong input, GETTOKEN_NOMEM = -2 for no memory available); in case of
 success, a pointer to a linked list of Token structures is also
 returned.  Every node in this linked list has been allocated on the
 heap using malloc(), and MUST be free()ed by the caller.

 "Null tokens", i.e. nothing and/or whitespace only between two
 separators (or between the beginning of the string and a separator,
 or between a separator and the end of the string), are allowed; in
 that case the related Token node in the linked list contains an empty
 string.

 getToken() is implemented as a finite state machine; the SM driver
 table is the following:

                  +-------------+---------------+------------+---------------+
                  |  end string |   delimiter   | whitespace |     other     |
 +----------------+-------------+---------------+------------+---------------+
 |    state 0     |    exit     | push null to- |  continue  | set start,end |
 |   (initial)    |             | ken; state=1  |            |   state = 2   |
 +----------------+-------------+---------------+------------+---------------+
 |    state 1     |  push null  | push null to- |  continue  | set start,end |
 | (out of token) | token; exit | ken; state=1  |            |   state = 2   |
 +----------------+-------------+---------------+------------+---------------+
 |    state 2     | push token; |  push token;  | push token |    set end    |
 | (inside token) |    exit     |   state = 1   | state = 3  |   state = 2   |
 +----------------+-------------+---------------+------------+---------------+
 | state 3 (after |    exit     |   state = 1   |  continue  | set start,end |
 | end of token)  |             |               |            |   state = 2   |
 +----------------+-------------+---------------+------------+---------------+

 However what is saved is not start and end of the token, but start
 and _length_ of the token.

 Usage example:

    char string[WHATEVER], *pc;
    int nTokens;
    Token *pT;
    ...
    if ((nTokens = getToken(string, NULL, &pT) < 0) {
      puts("Error return from getToken");
    } else {
      printf("%d tokens parsed from the given string\n", nTokens);
      while (pT) {
        Token *pN = pT->next;
        printf("- \"%s\"\n", pT->name);
        free(pT);
        pT = pN;
      }
    }

 $Id: gettoken.c,v 1.4 2000/10/24 09:26:12 loreti Exp $

 *---------------------------------------------------------*/

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "gettoken.h"

#define GETTOKEN_OK 0

static void freeList(Token *);
static int pushToken(const char *, const size_t, const size_t,
                     Token **, Token **, size_t *);

#ifdef TEST
#include <stdio.h>

int main()                      /* Quick and dirty. Sorry. */
{
  char string[128], sep[16], *pc, *psep;
  int nTokens;
  Token *pT;

  for (;;) {
    printf("String ('\\' to exit) ? ");
    fgets(string, 128, stdin);
    if ((pc = strchr(string, '\n')) != 0) *pc = '\0';
    if (!strcmp(string, "\\")) return 0;

    printf("Delimiters (or '\\' for none) ? ");
    fgets(sep, 16, stdin);
    if ((pc = strchr(sep, '\n')) != 0) *pc = '\0';
    psep = (strcmp(sep, "\\") ? sep : 0);

    if ((nTokens = getToken(string, sep, &pT)) < 0) {
      printf("Error return = %d from getToken\n", nTokens);
    } else {
      int i = 0;

      printf("%d tokens parsed from the given string\n", nTokens);
      while (pT) {
        Token *pN = pT->next;

        printf("%d : \"%s\"\n", ++i, pT->name);
        free(pT);
        pT = pN;
      }
    }

    printf("- THE END -\n");
  }
}
#endif /* TEST */

int getToken(
  const char  *string,          /* String to be parsed             */
  const char  *sep,             /* Separator characters, or NULL   */
  Token      **ppT              /* Pointer to the linked list root */
) {
  Token         *root    = 0;   /* First node in the linked list   */
  Token         *last    = 0;   /* Last node in the linked list    */
  size_t         nTokens = 0;   /* Number of nodes                 */
  size_t         tStart  = 0;   /* Offset of the current token     */
  size_t         tLen    = 0;   /* Length of the current token     */
  size_t         i       = 0;   /* Current offset in "string"      */
  unsigned char  c;             /* Current character in "string"   */
  int            state   = 0;   /* SM current status               */

  if (ppT) *ppT = 0;
  if (string == 0   ||   ppT == 0) return GETTOKEN_WRONG;

  while ((c = string[i++]) != '\0') {
    if (strchr(sep, c) != 0) {
      if (state != 3) {
        if (pushToken(string, tStart, tLen, &root, &last, &nTokens)) {
          freeList(root);
          return GETTOKEN_NOMEM;
        }
        tStart = tLen = 0;
      }
      state = 1;
    } else if (isspace(c)) {
      if (state == 2) {
        if (pushToken(string, tStart, tLen, &root, &last, &nTokens)) {
          freeList(root);
          return GETTOKEN_NOMEM;
        }
        tStart = tLen = 0;
        state  = 3;
      }
    } else {
      if (state != 2) {
        tStart = i - 1;
        state  = 2;
      }
      tLen = i - tStart;
    }
  }

  if (state == 1   ||   state == 2) {
    if (pushToken(string, tStart, tLen, &root, &last, &nTokens)) {
      freeList(root);
      return GETTOKEN_NOMEM;
    }
  }

  *ppT = root;
  return nTokens;
}

static void freeList(
  Token *pT
) {
  while (pT != 0) {
    Token *pN = pT->next;

    free(pT);
    pT = pN;
  }
}
static int pushToken(
  const char    *string,        /* String to be parsed                */
  const size_t   start,         /* Token offset                       */
  const size_t   length,        /* Token length                       */
  Token        **pRoot,         /* First node in the linked list      */
  Token        **pLast,         /* Last node in the linked list       */
  size_t        *pNum           /* Number of nodes in the linked list */
) {
  Token *pN;

  if ((pN = malloc(sizeof(Token) + length)) == 0) {
    return GETTOKEN_NOMEM;
  }

  pN->next = 0;
  if (length > 0) strncpy(pN->name, string+start, length);
  pN->name[length] = '\0';

  if (*pLast == 0) {
    *pRoot = *pLast = pN;
  } else {
    *pLast = (*pLast)->next = pN;
  }

  ++(*pNum);
  return GETTOKEN_OK;
}
